/********************************************************************************
*                                                                               *
*                  Generic Database Interface                                   *
*                                                                               *
*********************************************************************************
* Copyright (C) 2002 by Mathew Robertson.       All Rights Reserved.            *
* Copyright (C) 2002 by Giancarlo Formicuccia.  All Rights Reserved.            *
*********************************************************************************
* This library is free software; you can redistribute it and/or                 *
* modify it under the terms of the GNU Lesser General Public                    *
* License as published by the Free Software Foundation; either                  *
* version 2.1 of the License, or (at your option) any later version.            *
*                                                                               *
* This library is distributed in the hope that it will be useful,               *
* but WITHOUT ANY WARRANTY; without even the implied warranty of                *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU             *
* Lesser General Public License for more details.                               *
*                                                                               *
* You should have received a copy of the GNU Lesser General Public              *
* License along with this library; if not, write to the Free Software           *
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA.    *
*********************************************************************************/
#ifndef FXDATABASEINTERFACE_H
#define FXDATABASEINTERFACE_H

/**
 * FXDatabaseInterface is the first point of call when we want to create a real database
 * interface, such as an ODBC interface.  Child classes implement specific handling of
 * connecting to a real database, executing the real query, etc.
 */

#ifndef FXBASEOBJECT_H
#include "FXBaseObject.h"
#endif
namespace FXEX {
class FXDatabaseQuery;

/**
 * Base class for connection handling.
 * A class derived from FXDatabaseInterface must implement
 * at least all the pure virtual functions.
 */
class FXDatabaseInterface: public FXBaseObject {
  FXDECLARE_ABSTRACT(FXDatabaseInterface)
  friend class FXDatabaseQuery;

private:
  FXString database;  // name of database
  FXString username;  // user of database connection
  FXString password;  // password to database connection

protected:
  FXDatabaseHandle dbhandle;  // handle to actual database
  FXbool           connected; // connection status
  FXint            code;      // error code

protected:

  /// serialisation
  FXDatabaseInterface();

  /// ctor
  FXDatabaseInterface(FXApp *a,FXObject *tgt=NULL,FXSelector sel=0);

  /// children implement a real connect method
  virtual FXbool connect() { return FALSE; }

  /// children implement a real disconnect method
  virtual void disconnect() {}

  /// implement real query mechanism
  virtual FXbool query(const FXString& qry)=0;

public:

  /// get the status
  FXint status() { return code; }

  /// indicated whether the object is connected to a real database
  FXbool isConnected() const { return connected; }

  /// set the name of the database we want to be connected to
  void setDatabase(const FXString &name);

  /// get the name of the database we are connected to
  FXString getDatabase() const { return database; }

  /// set the name of the user we want to connect to the database as
  void setUser(const FXString &name);

  /// get the name of the user that we are connected to the database as
  FXString getUser() const { return username; }

  /// set the password for the user
  void setPassword(const FXString &passwd);

  /// this returns a list of tables implemented for this database
  virtual FXStringList getTables()=0;

  /**
   * execute an SQL query, which doesn't need to return any results,
   * such as an INSERT, APPEND, UPDATE
   */
  virtual FXbool execute(const FXString &command);

  /// prepare a query ready for execution
  FXDatabaseQuery* prepare(const FXString& query);

  /// create resources
  virtual void create();

  /// Stream save and load
  virtual void save(FXStream& store) const;
  virtual void load(FXStream& store);

  /// dtor
  virtual ~FXDatabaseInterface();
  };

} // namespace FXEX
#endif // FXDATABASEINTERFACE_H
