/********************************************************************************
*                                                                               *
*                  File I/O object                                              *
*                                                                               *
*********************************************************************************
* Copyright (C) 2002 by Mathew Robertson.   All Rights Reserved.                *
*********************************************************************************
* This library is free software; you can redistribute it and/or                 *
* modify it under the terms of the GNU Lesser General Public                    *
* License as published by the Free Software Foundation; either                  *
* version 2.1 of the License, or (at your option) any later version.            *
*                                                                               *
* This library is distributed in the hope that it will be useful,               *
* but WITHOUT ANY WARRANTY; without even the implied warranty of                *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU             *
* Lesser General Public License for more details.                               *
*                                                                               *
* You should have received a copy of the GNU Lesser General Public              *
* License along with this library; if not, write to the Free Software           *
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA.    *
********************************************************************************/
#ifndef FXFILEIO_H
#define FXFILEIO_H

#ifndef FXIOHANDLE_H
#include "FXIOHandle.h"
#endif
namespace FXEX {

/**
 * This class implements accessing data from files, including support for seeking
 * around the file, and support for read and write locking all or parts of a file.
 */

enum {
  FILEIO_NONE      = 0,
  FILEIO_EXCLUSIVE = 0x00000001,
  FILEIO_TRUNCATE  = 0x00000002,
  FILEIO_SYNC      = 0x00000004
  };

class FXAPI FXFileIO : public FXIOHandle {
  FXDECLARE(FXFileIO)

  public:
    enum{
      LockUnknown=-1,   // couldn't determine lock status
      Unlocked=0,       // not locked
      LockRead=1,       // this process is currently holding an read lock
      LockWrite=2,      // this process is currently holding an write lock
      LockedRead=3,     // read locked by another process
      LockedWrite=4,    // write locked by another process
      };

  protected:
    FXString filename;
    FXuint retryInterval;
    FXint locked;          // locked status
    void* region;          // mapped region
    FXuint mappedLength;   // length of mapped region

  protected:
    FXFileIO(){}

    /// instantiates this class
    virtual FXFileIO* newInstance(FXInputHandle h);

  public:
    /// get access to a file
    FXFileIO(FXApp *a,const FXString& file,FXObject *tgt=NULL,FXSelector sel=0,FXuint opts=FILEIO_NONE);

    /// use an already open file :-)
    FXFileIO(FXInputHandle f,FXApp *a,FXObject *tgt=NULL,FXSelector sel=0);

    /// create and open a temporary file, deletes on close
    FXFileIO(FXApp *a,FXObject *tgt=NULL,FXSelector=0);

    /// get the filename
    FXString getFilename() const;

    /// open the file
    virtual FXbool open();

    /// close the file
    virtual void close();

    /// get the current size
    FXlong size();

    /// what is current file pointer position
    FXlong current();

    /// move to the start of file plus some number of pos bytes
    FXbool seek(FXlong pos=0);

    /// move to some point in the file, from where it is currently
    FXbool seekcur(FXlong pos);

    /// move to the end of the file minus pos
    /// Note: zero is the first byte beyond the end of the file,
    ///       -1 is one past the end of the file (differs from lseek)
    FXbool seekend(FXlong pos=0);

    /// truncate the file to some length
    /// Note: you should really check the size of the file, after you call truncate
    FXbool truncate(FXlong len);

    /// nice and simple lock interface
    FXbool lock() { return writeLock(); }

    /// nice and simple unlock interface
    FXbool unlock(FXlong len);

    /// is the file already locked
    /// - a length of zero implies the whole file
    /// - for the simple interface, a non-zero return code implies some form of lock
    /// - otherwise, returns 'Unlocked' (0) if not locked, or the type of lock
    FXint isLocked(FXlong len=0);

    /// lock a file - read lock
    /// - multiple read locks can be in-force (by multiple processes) at any given time
    /// - never unlocks any other lock
    /// - locking with a length of zero means lock the whole file
    /// - otherwise lock the number of bytes, begining from the current position
    /// - timeout is in milliseconds
    FXbool readLock(FXlong len=0,FXint timeout=0);

    /// lock a file - write lock
    /// - only one lock can be in force at any given time, ie the write lock
    /// - cannot unlock any other lock
    /// - a write lock still allows you to read the file (ie it is a more powerful read lock)
    /// - locking with a length of zero means lock the whole file
    /// - otherwise lock the number of bytes, begining from the current position
    /// - timeout is in milliseconds
    FXbool writeLock(FXlong len=0,FXint timeout=0);
    
    /// this allows us to adjust or retry interval when trying to set a lock
    /// - retry in nanoseconds
    void setLockRetry(FXuint retry) { retryInterval=retry; }

    /// this class supports memory mapping of the file, but only one region at a time...
    /// load part of the file into memory, from the current seek position
    FXbool map(FXlong len,FXbool writeable=TRUE);

    /// unload part of a file from memory
    void unmap();

    /// is the file mapped into memory
    FXbool isMapped() { return region?TRUE:FALSE; }

    /// return pointer to mapped memory
    void* getMapping() { return region; }

    /// duplicate this file handle
    FXFileIO* duplicate(FXInputHandle newHandle=INVALID_HANDLE);

    /// save object to stream
    virtual void save(FXStream& store) const;

    /// load object from stream
    virtual void load(FXStream& store);

    /// dtor
    virtual ~FXFileIO();
  };

} // namespace FXEX
#endif // FXFileIO
