/********************************************************************************
*                                                                               *
*         FOX extension library: Definitions, Types, and Macros                 *
*                                                                               *
*********************************************************************************
* Copyright (C) 2002 by Mathew Robertson.   All Rights Reserved.                *
*********************************************************************************
* This library is free software; you can redistribute it and/or                 *
* modify it under the terms of the GNU Lesser General Public                    *
* License as published by the Free Software Foundation; either                  *
* version 2.1 of the License, or (at your option) any later version.            *
*                                                                               *
* This library is distributed in the hope that it will be useful,               *
* but WITHOUT ANY WARRANTY; without even the implied warranty of                *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU             *
* Lesser General Public License for more details.                               *
*                                                                               *
* You should have received a copy of the GNU Lesser General Public              *
* License along with this library; if not, write to the Free Software           *
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA.    *
********************************************************************************/
#ifndef FXEXDEFS_H
#define FXEXDEFS_H
 
// loads the FOX defs
#ifndef FXDEFS_H
#include <fox/fxdefs.h>
using namespace FX;
#endif

// Disable warnings on extern before template instantiation for the Microsoft compiler.
// see "HOWTO: Exporting STL Components Inside & Outside of a Class" on the microsoft website
#if defined(WIN32) && defined(_MSC_VER)
#pragma warning (disable : 4231)
#endif


// implement CALLBACK for unix
#ifndef CALLBACK
#define CALLBACK
#endif


/// 'endl' is useful if we are using streaming
#ifndef endl
#  define endl '\n'
#endif


namespace FXEX {

// provide an extern mechanism
#ifdef WIN32
#ifdef FOXDLL
#ifndef FOXDLL_EXPORTS
#define FXEXTERN extern
#endif
#endif
#endif

#ifndef FXEXTERN
#define FXEXTERN
#endif


// Valid compression factors for Bzip and Gzip compression libraries
#define COMPRESS_MAX 9
#define COMPRESS_NORMAL 6
#define COMPRESS_FAST 1
#define COMPRESS_NONE 0


// Win32 defines INFINITE to be -1, we might as well do it too.
#ifndef WIN32
# define INFINITE (-1)
#endif


// determine the newline charater(s)
#ifdef WIN32
#define FXNEWLINE "\r\n"
#endif
#ifdef MAC
#define FXNEWLINE "\r"
#endif
#ifndef FXNEWLINE
#define FXNEWLINE "\n"
#endif


// Make a long from two ints - provided longs are twice the size of an int
#ifdef FX_LONG
#  define MKLONG(l,h) ( (((FXlong)(l))&0xffffffff) | (((FXlong)(h))<<32) )
#else
#  if defined _FILE_OFFSET_BITS && _FILE_OFFSET_BITS == 64
#    error "You wanted 64bit file sizes (actually 63), but I couldn't make it so..."
#  else
#    warning "FX_LONG is not enabled for this platform..."
#    warning "This means that you are using 32bit longs, rather than 64bit..."
#    warning "=> assuming FXlong == long"
typedef unsigned long  FXulong;
typedef long           FXlong;
#  endif
#endif


// quite a while ago (18/08/2000), these macros (or something similar) were
// posted to the FOX users list, by Dirk Zoller. I cleaned them up a little.
#define FXDeclare(Class) \
    FXDECLARE(Class)
#define FXDeclareAbstract(Class) \
    FXDECLARE_ABSTRACT(Class)
#define FXDefMap(Class) \
    FXDEFMAP(Class) Class##Map[]
#define FXImplementAbstract(Class,Parent) \
    FXIMPLEMENT_ABSTRACT(Class,Parent,Class##Map,ARRAYNUMBER(Class##Map))
#define FXImplement(Class,Parent) \
    FXIMPLEMENT(Class,Parent,Class##Map,ARRAYNUMBER(Class##Map))
#define FXFuncDecl(Func) \
    long on##Func (FXObject*,FXSelector,void*)
#define FXFuncImpl(Class,Func,tgt,sel,ptr) \
    long Class::on##Func (FXOject *tgt,FXSelector sel, void *ptr)
#define FXMapTypes(Class,FromType,ToType,Func) \
    FXMAPTYPES(SEL_##FromType,SEL_##ToType,Class::on##Func)
#define FXMapType(Class,SelType,Func) \
    FXMAPTYPE(SEL_##SelType,Class::on##Func)
#define FXMapFuncs(Class,SelType,FromId,ToId,Func) \
    FXMAPFUNCS(SEL_##SelType,Class::ID_##FromId,Class::ID_##ToId,Class::on#Func)
#define FXMapFunc(Class,SelType,Id,Func) \
    FXMAPFUNC(SEL_##SelType,Class::ID_##Id,Class::on##Func)


// Zed A Shaw posted these (09/09/02), or a variation of them
#define FXSEND(tgt,sel,msg,ptr) \
    (tgt->handle(this,MKUINT(msg,sel),ptr)
#define FXROUTE(src,tgt,sel,msg,ptr) \
    (tgt->handle(src,MKUINT(msg,sel),ptr)


// debugging macros
#ifndef NDEBUG
#  define FXCTOR() fxtrace (100,"%s::%s %p\n",getClassName(),getClassName(),this)
#  define FXMETHOD(methodname) fxtrace (100,"%s::%s %p\n",getClassName(),#methodname,this)
#  define FXDTOR() fxtrace (100,"%s::~%s %p\n",getClassName(),getClassName(),this)
#else
#  define FXCTOR()
#  define FXMETHOD(methodname)
#  define FXDTOR()
#endif

     
// New selector types
enum {
  SEL_DATA=FX::SEL_LAST,      // form of data as an event
  SEL_THREAD,                 // thread/runnable events
  SEL_WAITABLE_WAIT,          // waitable event such as a Condition variable, semaphore, etc
  SEL_WAITABLE_ACTIVATE,      // waitable event such as a Condition variable, semaphore, etc
  SEL_INTERLOCK,              // interlocked event; object went into lock
  SEL_BARRIER_LOCK,           // barrier event; thread waiting in barrier lock
  SEL_BARRIER_UNLOCK,         // barrier event; barrier object reset, threads released
  SEL_INPUT,                  // some type of input event
  SEL_OUTPUT,                 // some type of output event
  SEL_ERROR,                  // some type of error event
  SEL_IO,                     // Some form of IO
  SEL_EVENT,                  // a generic event
  SEL_BEGIN,                  // en event defining some begining
  SEL_END,                    // an event defining some ending
  SEL_TAG,                    // tag event
  SEL_CONTENT,                // content event
  SEL_REGISTRY,               // a registry event (TODO I have a plan for this...)
  SEL_LASTEX                  // Last message
  };


// IO status definitions
typedef FXint FXIOStatus;
enum {
  FXIOStatusUnknown=-2,
  FXIOStatusError=-1,
  FXIOStatusOk=0,
  FXIOStatusOK=0,
  FXIOStatusLast
  };
#ifndef INVALID_HANDLE
#  ifdef WIN32
#    define INVALID_HANDLE INVALID_HANDLE_VALUE
#  else
#    define INVALID_HANDLE -1
#  endif
#endif


// IO state definitions
typedef FXint FXIOState;
enum {
  FXIOStateUnknown=-1,
  FXIOStateNone=-1,
  FXIOStateOk=0,
  FXIOStateOK=0,
  FXIOStateUnconnected,
  FXIOStateConnected,
  FXIOStateOpen=FXIOStateConnected,
  FXIOStateListener,
  FXIOStateAccepted,
  FXIOStateDuplicated,
  FXIOStateLast
  };


// socket stuff
typedef FXint FXport;


/*
// need size_t in lots of places - if its not defined, define it...
#ifndef __size_t
typedef unsigned int		size_t;
#endif
*/

// thread stuff
#ifndef WIN32
#include <pthread.h>    // sigh - there has to be a better way to do this
#include <semaphore.h>  // sigh - again
typedef pthread_t*		FXThreadHandle;
typedef pthread_mutex_t		FXThreadMutex;
typedef pthread_cond_t		FXThreadCondition;
typedef FXInputHandle*		FXThreadEventHandle;
typedef sem_t			FXThreadSemaphore;
#else
typedef HANDLE			FXThreadHandle;
typedef HANDLE		 	FXThreadMutex;
typedef HANDLE			FXThreadCondition;
typedef HANDLE			FXThreadSemaphore;
typedef FXInputHandle		FXThreadEventHandle;
#endif


// dynamic library loading
#ifndef WIN32
typedef void*			FXDLLHandle;
#else
typedef HMODULE			FXDLLHandle;
#endif


// database interface handle
typedef void*                   FXDatabaseHandle;


namespace FXUtils {

////////////// global functions ////////////////////


/// time conversion routines
#ifdef WIN32
static void get_time_now(unsigned long* abs_sec, unsigned long* abs_nsec);
static DWORD get_timeout(unsigned long secs, unsigned long nsecs,DWORD default_to);
#else
void convert_timeval(struct timeval *tv,FXuint ms);
void convert_timespec(struct timespec *ts,FXuint ms);
#endif

/// save the machine dependant endianness value to the stream
extern FXAPI void saveEndianness(FXStream& store);

/// load the machine dependant endianness value from the stream,
/// and setup stream to swap the bytes if the stream byte order
/// is non-native to the CPU architecture
extern FXAPI void loadEndianness(FXStream& store);

/// create a random seed for use with fxrandom
/// it is based on the current time, so its a _reasonably_ random seed
extern FXAPI FXuint fxrandomseed();

/// a useful function call for generating random numbers
/// generate an integer/float number [lowerBound,upperBound)
extern FXAPI FXint fxrand(FXint upperBound,FXint lowerBound=0);
extern FXAPI FXfloat fxrand(FXfloat upperBound,FXfloat lowerBound=0.0);

////////////// global variables ////////////////////

/// Version number that the library version is compiled with
extern FXAPI const FXuchar fxexversion[3];

}; // namespace FXUtils
}; // namespace FXEX

#endif // FXEXDEFS_H

