/********************************************************************************
*                                                                               *
*               Dynamically linked library loader (DLL loader)                  *
*                                                                               *
*********************************************************************************
* Copyright (C) 2002 by Mathew Robertson.   All Rights Reserved.                *
*********************************************************************************
* This library is free software; you can redistribute it and/or                 *
* modify it under the terms of the GNU Lesser General Public                    *
* License as published by the Free Software Foundation; either                  *
* version 2.1 of the License, or (at your option) any later version.            *
*                                                                               *
* This library is distributed in the hope that it will be useful,               *
* but WITHOUT ANY WARRANTY; without even the implied warranty of                *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU             *
* Lesser General Public License for more details.                               *
*                                                                               *
* You should have received a copy of the GNU Lesser General Public              *
* License along with this library; if not, write to the Free Software           *
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA.    *
********************************************************************************/
#include <fox/fxver.h>
#include <fox/xincs.h>
#include <fox/fxdefs.h>
#include <fox/FXStream.h>
#include <fox/FXString.h>
#include <fox/FXFile.h>
using namespace FX;
#include "exincs.h"
#include "fxexdefs.h"
#include "FXDLL.h"
using namespace FXEX;
namespace FXEX {

typedef void (*FXDLL_func)();

// ctor
FXDLL::FXDLL(const FXString& filename){
  file=filename;
  handle=NULL;
  refcount=0;
  }

// dtor - unload library on delete if necessary
FXDLL::~FXDLL(){
  while (refcount) close();
  handle=(FXDLLHandle)-1;
  }

// opens the file using the OS specific implementation
FXDLLHandle FXDLL::loadLibrary(const FXString& filename){
#ifndef WIN32
  return ::dlopen(filename.text(),RTLD_NOW|RTLD_GLOBAL);
#else
  return LoadLibrary(filename.text());
#endif
  }

// closes the library using the OS specific implementation
FXDLLHandle FXDLL::unloadLibrary(){
#ifndef WIN32
  ::dlclose(handle);
#else
  FreeLibrary(handle);
#endif
  return NULL;
  }

// set to new filename...
// if the file is the same as previous, then just increment the refcount
// otherwise the file is not the same name as previous, so if a refcount
// exists, then we cannot open any more objects
void FXDLL::setFilename(const FXString& filename){
  if (file == filename) refcount++;
  else if (refcount) fxerror("FXDLL: cannot change filename while library loaded\n");
  file=filename;
  }

// open the library file
FXbool FXDLL::open(){
  FXString name=FXFile::name(file);
  FXString path=FXFile::directory(file);
  if (path.length()) path += PATHSEP;
#ifndef WIN32
  FXString pathenv=FXFile::getEnvironment("LD_LIBRARY_PATH");
  FXString extalt=".so";
#else
  FXString pathenv=FXFile::getEnvironment("PATH");
  FXString extalt=".dll";
#endif

  // this is a pretty crude way of selecting functionality, but it is
  // reasonably easy to read... and scales reasonbly easily...
  // try the normal filename, then we try alternative library names,
  // ie libXXX, XXX.lib, XXX.dll, etc
  FXString f;
  FXint state=0;
  while (state>0) {
    switch (state++) {
      case 0: f=file; break;
      case 1: f=path + name + extalt; break;
      case 2: f=path + "lib" + name + extalt; break;
      case 3: f=path + name + "dll" + extalt; break;
      case 4: f=path + name + ".lib"; break;
      default: return FALSE;
      }
    if ((handle=loadLibrary(f))!=NULL) {
      refcount++;
      FXDLL_func func = (FXDLL_func) getFunction("FXDLL_load");
      if (func!=NULL) (*func)();
      return TRUE;
      }
    }
  return FALSE;
  }

// unload the library
void FXDLL::close(){
  if (!refcount) return;
  refcount--;
  if (!refcount) {
    FXDLL_func func = (FXDLL_func) getFunction("FXDLL_unload");
    if (func!=NULL) (*func)();
    handle=unloadLibrary();
    }
  }

// indicate if the library is already loaded
FXbool FXDLL::isOpen(){
  return refcount?TRUE:FALSE;
  }

// return a pointer to a function in the library
void* FXDLL::getFunction(const FXString& function){
#ifndef WIN32
  FXDLLHandle h=::dlsym(handle,function.text());
  if ( dlerror() != NULL ) return NULL;
  return h;
#else
  return GetProcAddress(handle.function.text());
#endif
  }

}

