/********************************************************************************
*                                                                               *
*                  File I/O object                                              *
*                                                                               *
*********************************************************************************
* Copyright (C) 2002 by Mathew Robertson.   All Rights Reserved.                *
*********************************************************************************
* This library is free software; you can redistribute it and/or                 *
* modify it under the terms of the GNU Lesser General Public                    *
* License as published by the Free Software Foundation; either                  *
* version 2.1 of the License, or (at your option) any later version.            *
*                                                                               *
* This library is distributed in the hope that it will be useful,               *
* but WITHOUT ANY WARRANTY; without even the implied warranty of                *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU             *
* Lesser General Public License for more details.                               *
*                                                                               *
* You should have received a copy of the GNU Lesser General Public              *
* License along with this library; if not, write to the Free Software           *
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA.    *
********************************************************************************/
#ifndef _POSIX_MAPPED_FILES
#define _POSIX_MAPPED_FILES
#endif
#include <fox/fxver.h>
#include <fox/xincs.h>
#include <fox/fxdefs.h>
#include <fox/FXStream.h>
#include <fox/FXString.h>
#include <fox/FXSize.h>
#include <fox/FXPoint.h>
#include <fox/FXRectangle.h>
#include <fox/FXRegistry.h>
#include <fox/FXApp.h>
#include <fox/FXObjectList.h>
#include <fox/FXFile.h>
using namespace FX;
#include "exincs.h"
#include "fxexdefs.h"
#include "FXFileIO.h"
using namespace FXEX;
namespace FXEX {

/* Note: we prefer to use fcntl rather than flock/lockf since each vender may implement
 *       those function calls slightly differently
 */

#define RETRY_INTERVAL 1000000       // 1 second

FXIMPLEMENT(FXFileIO,FXIOHandle,NULL,0)


// open file using specific filename
FXFileIO::FXFileIO(FXApp *a,const FXString& file,FXObject *tgt,FXSelector sel,FXuint opts) : FXIOHandle(a,tgt,sel) {
  filename=file;
  retryInterval=RETRY_INTERVAL;
  options&=opts;
  region=NULL;
  if(file.length()) code=FXIOStatusOk;
  }

// use pre-exisitng open'ed file
FXFileIO::FXFileIO(FXInputHandle f,FXApp *a,FXObject *tgt,FXSelector sel) : FXIOHandle(f,a,tgt,sel){
  retryInterval=RETRY_INTERVAL;
  region=NULL;
  if (f!=INVALID_HANDLE) {
    state=FXIOStateOk;
    code=FXIOStatusOk;
    }
  }

// open a temporary file
FXFileIO::FXFileIO(FXApp *a,FXObject *tgt,FXSelector sel) : FXIOHandle(a,tgt,sel){
  filename="....";
  retryInterval=RETRY_INTERVAL;
  region=NULL;
  code=FXIOStatusOk;
  }

FXFileIO::~FXFileIO(){}

void FXFileIO::save(FXStream &store) const {
  FXIOHandle::save(store);
  store << filename;
  store << retryInterval;
  }

void FXFileIO::load(FXStream &store) {
  FXIOHandle::load(store);
  store >> filename;
  store >> retryInterval;
  }

// return the filename
FXString FXFileIO::getFilename() const {
  return (filename.left(4)!="....")?(filename):(filename.right(filename.length()-4));
  }

// open the file
FXbool FXFileIO::open() {
  // make sure we haven't already opened a file,
  // see what the parents says about our handle...
  if (state!=FXIOStateNone) return FALSE;
  if (!FXIOHandle::open()) return FALSE;

  // open the file
  if (filename=="...."){
    FXString temp=FXFile::getTempDirectory() + PATHSEP + "XXXXXX";
    FXchar t[1024];
    strcpy(t,temp.text());
    filename+=t;
#ifndef WIN32
    iohandle=::mkstemp(t);
#else
#error TODO
#endif
    }
  else {
#ifndef WIN32
    FXint mode= O_RDWR|O_CREAT;
    mode |= options&FILEIO_EXCLUSIVE ? O_EXCL : 0;
    mode |= options&FILEIO_TRUNCATE ? O_TRUNC : 0;
    mode |= options&FILEIO_SYNC ? O_SYNC : 0;
    iohandle=::open(filename.text(),mode);
#else
    DWORD mode  = GENERIC_READ|GENERIC_WRITE;
    DWORD share = options&FILEIO_EXCLUSIVE ? 0 : FILE_SHARE_READ|FILE_SHARE_WRITE;
    DWORD trunc = options&FILEIO_TRUNCATE ? CREATE_ALWAYS : OPEN_ALWAYS;
    DOWRD sync  = options&FILEIO_SYNC ? SYNCHRONIZE : 0;
    iohandle=CreateFile(filename.text(),sync&mode,share,NULL,trunc,NULL);
#endif
    }
  if (iohandle==INVALID_HANDLE) {
    code=errno;  // FIXME on win32 this shoul be GetError or something 
    return FALSE;
    }
  FXTRACE((100,"file opened"));
  state=FXIOStateOpen;
  return TRUE;
  }

// delete the file if open as temp
void FXFileIO::close(){
  FXIOHandle::close();
  if (filename.left(4)=="....") FXFile::remove(getFilename());
  }

// file size
FXlong FXFileIO::size(){
  return (FXlong) FXFile::size(getFilename());
  }

// report the current file pointer position
FXlong FXFileIO::current(){
#ifndef WIN32
  return ::lseek(iohandle,0,SEEK_CUR);
#else
  return SetFilePointer(iohandle,0,0,FILE_CURRENT);
#endif
  }

// move around file, wrt the start of the file
FXbool FXFileIO::seek(FXlong pos){
  if (pos<0) fxerror("%s: cannot seek beyond start of file\n",getClassName());
#ifndef WIN32
  return ::lseek(iohandle,pos,SEEK_SET)>=0?TRUE:FALSE;
#else
  return SetFilePointer(iohandle,pos,NULL,FILE_BEGIN)>=0?TRUE:FALSE;
#endif
  }

// move around file, wrt the current position within the file
FXbool FXFileIO::seekcur(FXlong pos){
#ifndef WIN32
  return ::lseek(iohandle,pos,SEEK_CUR)>=0?TRUE:FALSE;
#else
  return SetFilePointer(iohandle,pos,NULL,FILE_CURRENT)>=0?TRUE:FALSE;
#endif
  }

// move around file, wrt the end of the file
FXbool FXFileIO::seekend(FXlong pos){
#ifndef WIN32
  return ::lseek(iohandle,pos,SEEK_END)>=0?TRUE:FALSE;
#else
  return SetFilePointer(iohandle,pos,NULL,FILE_END)>=0?TRUE:FALSE;
#endif
  }

// get lock status
FXint FXFileIO::isLocked(FXlong len){
#ifndef WIN32
  struct flock lock;
  lock.l_type=F_WRLCK;
  lock.l_start=current();
  lock.l_whence=SEEK_CUR;
  lock.l_len=len;
  if (::fcntl(iohandle,F_GETLK,&lock)<0) return LockUnknown;

  // handle unlocked case
  if (lock.l_type & F_UNLCK) return Unlocked;

  // handle 'this process' case
  // (first test for write, then for read)
  if (lock.l_pid == fxgetpid()) {
    if (lock.l_type & F_WRLCK) return LockWrite;
    else return LockRead;
    }

  // handle 'some other process case'
  // (first test for write, then for read)
  if (lock.l_type & F_WRLCK) return LockedWrite;
  else return LockedRead;

#else

  // FIXME implement test for file locking - need to support Read and Write
  WINBOOL result= LockFileEx(iohandle,LOCKFILE_FAIL_IMMEDIATELY,0,(DWORD)len,0,NULL);
  if (result) UnlockFileEx(iohandle,9,(DWORD)len,0,NULL);
  return result?Unlocked:LockedWrite;
#endif
  }

// set a read lock
FXbool FXFileIO::readLock(FXlong len,FXint timeout){
  // see if we already have a lock
  FXint locking=isLocked(len);
  if (locking == LockRead || locking == LockWrite) return TRUE;
  // there is no lock on the file, or some other process has a lock,
  // just try and continually lock the file until we timeout
 
#ifndef WIN32
  struct flock lock;
  lock.l_type=F_RDLCK;
  lock.l_whence=SEEK_CUR;
  lock.l_start=current();
  lock.l_len=len;
  if (timeout==0) return ::fcntl(iohandle,F_SETLK,&lock)<0?FALSE:TRUE;
  if (timeout<0) {
    while (::fcntl(iohandle,F_SETLK,&lock)<0) {}
    return TRUE;
    }
  register FXint to=timeout*1000000;
  while(to>0){
    if (::fcntl(iohandle,F_SETLK,&lock)>-1) return TRUE;
    fxsleep(retryInterval);
    to-=retryInterval;
    }
  return FALSE;
#else
  if (timeout==0) return LockFileEx(iohandle,LOCKFILE_FAIL_IMMEDIATELY,0,(DWORD)len,0,NULL)?TRUE:FALSE;
  if (timeout<0) return LockFileEx(iohandle,0,0,(DWORD)len,0,NULL)?TRUE:FALSE; // FIXME check
  register FXint to=timeout*1000000;
  while(to>0){
    if (LockFileEx(iohandle,LOCKFILE_FAIL_IMMEDIATELY,0,(DWORD)len,0,NULL)) return TRUE;
    fxsleep(retryInterval);
    to-=retryInterval;
    }
  return FALSE;
#endif
  }

// set a write lock
FXbool FXFileIO::writeLock(FXlong len,FXint timeout){

  // see if we already have a lock
  FXint locking=isLocked(len);
  if (locking == LockWrite) return TRUE;

  // there is no lock on the file, or some other process has a lock,
  // just try and continually lock the file until we timeout

#ifndef WIN32
  struct flock lock;
  lock.l_type=F_WRLCK;
  lock.l_whence=SEEK_CUR;
  lock.l_start=current();
  lock.l_len=len;
  if (timeout==0) return ::fcntl(iohandle,F_SETLKW,&lock)<0?FALSE:TRUE;
  if (timeout<0) {
    while (::fcntl(iohandle,F_SETLKW,&lock)<0) {}
    return TRUE;
    }
  register FXint to=timeout*1000000;
  while(to>0){
    if (::fcntl(iohandle,F_SETLKW,&lock)>-1) return TRUE;
    fxsleep(retryInterval);
    to-=retryInterval;
    }
  return FALSE;
#else

  if (timeout==0) return LockFileEx(iohandle,LOCKFILE_FAIL_IMMEDIATELY,0,(DWORD)len,0,NULL)?TRUE:FALSE;
  if (timeout<0) return LockFileEx(iohandle,0,0,(DWORD)len,0,NULL)?TRUE:FALSE; // FIXME check
  register FXint to=timeout*1000000;
  while(to>0){
    if (LockFileEx(iohandle,LOCKFILE_FAIL_IMMEDIATELY,0,(DWORD)len,0,NULL)) return TRUE;
    fxsleep(retryInterval);
    to-=retryInterval;
    }
  return FALSE;
#endif
  }

// unlock the last lock
FXbool FXFileIO::unlock(FXlong len){
#ifndef WIN32
  struct flock lock;
  lock.l_type=F_UNLCK;
  lock.l_start=current();
  lock.l_len=len;
  return ::fcntl(iohandle,F_SETLKW,&lock)<0?FALSE:TRUE;
#else
  return UnlockFileEx(iohandle,0,(DWORD)len,0)?TRUE:FALSE;
#endif
  }

// truncate the file to a specified length
FXbool FXFileIO::truncate(FXlong len){
  FXlong s=(FXlong)size();
  if (s == len) return TRUE;
  if (!seek(len)) return FALSE;  // set file pointer to new EOF position
  if (s < len) return TRUE;  // if the origonal size is less than the length, nothing left to do
#ifndef WIN32
  return ::ftruncate(iohandle,len)>=0?TRUE:FALSE;
#else
  return SetEndOfFile(iohandle)?TRUE:FALSE;
#endif
  }

// memory map a file
FXbool FXFileIO::map(FXlong len,FXbool writable){
  if (!len) { unmap(); return TRUE; }
#ifndef WIN32
  FXint mode= writable ? MAP_SHARED : MAP_PRIVATE;
  region=::mmap(0,(size_t)len,PROT_EXEC|PROT_READ|PROT_WRITE,mode,iohandle,current());
  if (region != MAP_FAILED) return TRUE;
  code=errno;
#else
  region=MapViewOfFile(iohandle,FILE_MAP_ALL_ACCESS,0,(DWORD)current(),(DWORD)len);
  if (region) return TRUE;
#endif
  region=NULL;
  return FALSE;
  }

// un-memory map a file
void FXFileIO::unmap(){
  if (!region) fxerror("%s: no region mapped\n",getClassName());
#ifndef WIN32
  ::munmap(region,mappedLength);
#else
  UnmapViewOfFile(region);
#endif
  region=NULL;
  }

/*
 * Create a file handle which is effectively a duplicate of this one
 */
FXFileIO* FXFileIO::duplicate(FXInputHandle newHandle){
#ifndef WIN32
  return dynamic_cast<FXFileIO*>(FXIOHandle::duplicate(newHandle));
#else
#error FIXME
#endif
  }

// implement the instance creation
FXFileIO* FXFileIO::newInstance(FXInputHandle h){
  return new FXFileIO(h,getApp(),target,message);
  }

}

