/********************************************************************************
*                                                                               *
*                  Monitor the status of a file                                 *
*                                                                               *
*********************************************************************************
* Copyright (C) 2002 by Mathew Robertson.   All Rights Reserved.                *
*********************************************************************************
* This library is free software; you can redistribute it and/or                 *
* modify it under the terms of the GNU Lesser General Public                    *
* License as published by the Free Software Foundation; either                  *
* version 2.1 of the License, or (at your option) any later version.            *
*                                                                               *
* This library is distributed in the hope that it will be useful,               *
* but WITHOUT ANY WARRANTY; without even the implied warranty of                *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU             *
* Lesser General Public License for more details.                               *
*                                                                               *
* You should have received a copy of the GNU Lesser General Public              *
* License along with this library; if not, write to the Free Software           *
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA.    *
********************************************************************************/
#include <fox/fxver.h>
#include <fox/xincs.h>
#include <fox/fxdefs.h>
#include <fox/FXStream.h>
#include <fox/FXString.h>
#include <fox/FXSize.h>
#include <fox/FXPoint.h>
#include <fox/FXRectangle.h>
#include <fox/FXRegistry.h>
#include <fox/FXApp.h>
#include <fox/FXFile.h>
using namespace FX;
#include "FXFileMonitor.h"
using namespace FXEX;
namespace FXEX {

FXDEFMAP(FXFileMonitor) FXFileMonitorMap[]={
  FXMAPFUNC(SEL_TIMEOUT,FXFileMonitor::ID_FILE_CREATE,FXFileMonitor::onFileCreate),
  FXMAPFUNC(SEL_TIMEOUT,FXFileMonitor::ID_FILE_CHANGED,FXFileMonitor::onFileChanged),
  FXMAPFUNC(SEL_TIMEOUT,FXFileMonitor::ID_FILE_END,FXFileMonitor::onFileEnd),
  };
FXIMPLEMENT(FXFileMonitor,FXBaseObject,FXFileMonitorMap,ARRAYNUMBER(FXFileMonitorMap))

// ctor
FXFileMonitor::FXFileMonitor(FXApp *a,const FXString& file,FXint ms,FXint to,FXObject *tgt,FXSelector sel) : FXBaseObject(a,tgt,sel) {
  filename=FXFile::absolute(file);
  if ((to*1000) < ms) fxerror("%s: end timeout must be bigger than scan time\n",getClassName());
  growtime=ms;
  endtimeout=to;
  growtimer=NULL;
  endtimer=NULL;
  size=0;
  }

// free up all resources
FXFileMonitor::~FXFileMonitor(){
  if (growtimer) growtimer=getApp()->removeTimeout(growtimer);
  if (endtimer) endtimer=getApp()->removeTimeout(endtimer);
  }

// create resources
void FXFileMonitor::create(){
  FXBaseObject::create();
  reset();
  }

// save object to stream
void FXFileMonitor::save(FXStream& store) const {
  FXBaseObject::save(store);
  store << filename;
  store << growtime;
  store << endtimeout;
  }

// load object from stream
void FXFileMonitor::load(FXStream& store){
  FXBaseObject::load(store);
  store >> filename;
  store >> growtime;
  store >> endtimeout;
  }

// reset the monitor
void FXFileMonitor::reset(){
  size=0;
  if (growtimer) growtimer=getApp()->removeTimeout(growtimer);
  if (endtimer) endtimer=getApp()->removeTimeout(endtimer);
  onFileCreate(NULL,0,NULL);
  }

// check if the file exists
long FXFileMonitor::onFileCreate(FXObject*,FXSelector,void*){
  if (! FXFile::exists(filename))
    growtimer=getApp()->addTimeout(growtime,this,ID_FILE_CREATE);
  else{
    growtimer=NULL;
    if (target) target->handle(this,FXSEL(SEL_CREATE,message),NULL);
    onFileChanged(NULL,0,NULL);
    }
  return 1;
  }

// generate a SEL_CHANGED event whenever the file grows or it mtime changes
long FXFileMonitor::onFileChanged(FXObject*,FXSelector,void*){
  if (FXFile::exists(filename)){
    FXTime mtime=FXFile::modified(filename);
    unsigned long sz=FXFile::size(filename);
    if (modified!=mtime || size!=sz){
      modified=mtime;
      size=sz;
      if (endtimer) endtimer=getApp()->removeTimeout(endtimer);
      if (target) target->handle(this,FXSEL(SEL_CHANGED,message),(void*)&sz);
      }
    else if (!endtimer && endtimeout!=-1)
      endtimer=getApp()->addTimeout(endtimeout*1000-growtime,this,ID_FILE_END);
    growtimer=getApp()->addTimeout(growtime,this,ID_FILE_CHANGED);
    }
  else{
    if (target) target->handle(this,FXSEL(SEL_DELETE,message),(void*)&size);
    reset();
    }
  return 1;
  }

// detect if a file hasn't changed size in a while
long FXFileMonitor::onFileEnd(FXObject*,FXSelector,void*){
  endtimer=NULL;
  growtimer=getApp()->removeTimeout(growtimer);
  if (target) target->handle(this,FXSEL(SEL_CLOSED,message),(void*)&size);
  growtimer=getApp()->addTimeout(growtime,this,ID_FILE_CHANGED);
  return 1;
  }

}

